# -*- coding: utf-8 -*-
from django.test import TestCase
from .calc_values import children_range, parents_list

class CalcValuesTests(TestCase):

    def test_children_range(self):

        self.assertEqual(children_range(27,4,2), [36, 53])
        self.assertEqual(children_range(36,4,2), [39, 44])
        self.assertEqual(children_range(45,4,2), [48, 53])
        self.assertEqual(children_range(54,4,2), [63, 80])
        self.assertEqual(children_range(78,4,2), [79, 80])

        self.assertEqual(children_range(-16,3,3), [-12, -1])
        self.assertEqual(children_range(-8,3,3), [-7, -5])
        self.assertEqual(children_range(16,3,3), [20, 31])
        self.assertEqual(children_range(28,3,3), [29, 31])

    def test_parents_list(self):
        
        self.assertEqual(parents_list(80,4,2), [54,72,78])
        self.assertEqual(parents_list(70,4,2), [54,63,69])
        self.assertEqual(parents_list(67,4,2), [54,63,66])

        self.assertEqual(parents_list(31,3,3), [16,28])
        self.assertEqual(parents_list(5,3,3), [0,4])





# -*- coding: utf-8 -*-
def children_range (id_e, tree_lv, tree_ch):
    y = tree_ch+1
    lv = tree_lv
    # Найдем уровень элемента
    for i in range(1, tree_lv):
        if (id_e % (y**(tree_lv-i))) == 0:
            lv = i
            break
    if lv == tree_lv: 
        return[]
    # Теперь вернем диапазон возможных детей (первого и последнего)
    return [id_e+y**(tree_lv-lv-1), id_e-1+((y**(tree_lv-lv-1))*(tree_ch+1))]

def parents_list(id_e, tree_lv, tree_ch):
    pl = []
    y = tree_ch+1
    for i in range(1, tree_lv):
        x = y**(tree_lv-i)
        e = x * (id_e // x)
        if e != id_e:
            pl.append(e)
    return pl



#Вариант с уровнем элемента:
#def parents_list(id_e, level, tree_lv, tree_ch):
#    pl = []
#    for i in range(1, level):
#        x = (tree_ch+1)**(tree_lv-i)
#        pl.insert(0, x * (id_e // x))
#    return pl




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

def load_stores_from_sql():
    #from django.conf import settings
    import os
    import treensl
    #sql_scripts = open(os.path.join(os.getcwd(),'treensl/sql_scripts/ini_postgres.sql'), 'r').read()
    #ini_sql = open(os.path.join(settings.BASE_DIR,'treensl/sql_scripts/ini_postgres.sql'), 'r').read()
    #ini_sql = open(os.path.join(treensl.__path__[0],'sql_scripts/ini_postgres.sql' ), 'r').read()
    ini_sql = open(os.path.join(treensl.__path__[0],
                                'sql_scripts/ini_postgres.sql'
                                ), 'r', newline='', encoding='utf-8'
                   ).read()
    return ini_sql

class Migration(migrations.Migration):

    dependencies = [

    ]

    operations = [
        migrations.RunSQL(load_stores_from_sql()),
    ]



# -*- coding: utf-8 -*-
from django.db import models

# Этот скрипт необходимо выполнять во всех иерархических таблицах-наследниках
# путем добавления в миграции вашего приложения migrations.RunSQL()
# Пример в myapp/migranions/0001_initial.py
POSTGRESQL_CONNECT_TABLE = '''-- тестировал в БД PostgreSQL 9.1 и выше

-- Добавляем поле-массив которое хранит "дырки"
ALTER TABLE  {0} ADD COLUMN holes integer[];

-- Начальный элемент дерева
INSERT INTO {0} ({1}) VALUES {2};

-- Имя таблицы и Размерность нового дерева
INSERT INTO tree_size (table_name, number_of_levels, number_of_children)
               VALUES ('{0}', {3}, {4});

-- Теперь можно подключить триггерные функции
CREATE TRIGGER after_upd
  AFTER UPDATE
  ON {0}
  FOR EACH ROW
  EXECUTE PROCEDURE {5};

CREATE TRIGGER before_del
  BEFORE DELETE
  ON {0}
  FOR EACH ROW
  EXECUTE PROCEDURE {6};

CREATE TRIGGER before_new
  BEFORE INSERT
  ON {0}
  FOR EACH ROW
  EXECUTE PROCEDURE {7};
'''


class TreeAbstract(models.Model):
    
    lvl = models.IntegerField(blank=False, null=False)
    created_children = models.IntegerField(blank=False, null=False)
    removed_children = models.IntegerField(blank=False, null=False)

    # Количество детей
    @property
    def count_children(self):
        return self.created_children - self.removed_children

    class Meta:
        abstract = True

    def sql_end(self):

        # В определении модели, поле 'parent_id',названо как 'parent'
        # Но, так как это поле ссылка на 'id', Django в БД его пишет как 'parent_id'
        # Создадим списки полей абстрактой модели и их значений для первой строки (1-й элемент)
        # (1-й элемент в таблице, хранящей дерево, ссылается сам на себя)
        list_fields = ['id', 'parent_id', 'lvl', 'created_children',
                       'removed_children']
                    #'removed_children', 'label_node']
        #list_values = [self.ROOT_ID, self.ROOT_ID, 0, 0, 0,'root']
        list_values = [self.ROOT_ID, self.ROOT_ID, 0, 0, 0]
        # Наследник модели может иметь свои дополнительные поля, и они могут быть NOT NULL
        # Необходимо вставить начальный элемент дерева, с учетом этого
        # То есть, заполнить такие поля каким-либо значением
        # (0 - подходит для полей всех типов в Postgres... вроде бы)
        for field in self._meta.fields:
            if (field.name not in list_fields) and (field.name != 'parent'):
                list_fields.append(field.name)
                list_values.append(0)
 
        p5 = 'treensl_after_upd()'
        p6 = 'treensl_before_del()'
        p7 = 'treensl_before_new()'

        return POSTGRESQL_CONNECT_TABLE.format(self._meta.db_table,     #Название таблицы-наследника в БД
                                 ", ".join(list_fields,),
                                 tuple(list_values),
                                 self.LEVELS,
                                 self.CHILDREN,
                                 p5, p6, p7)

class Tree32Abstract(TreeAbstract):
    # Некоторые сочетания уровень/дети для int32:
    # 3/1624, 4/255, 5/83, 6/39, 7/23, 8/15
    # Размерность дерева по умолчанию:
    LEVELS = 5
    CHILDREN = 83

    # id корня дерева 5/83
    ROOT_ID = -2091059712

    id = models.IntegerField(primary_key=True)
    parent = models.ForeignKey('self')
    
    class Meta:
        abstract = True



class Tree64Abstract(TreeAbstract):
    # Некоторые сочетания уровень/дети для int64:
    # 3/2642245, 4/65535, 5/7131, 6/1624, 7/564, 8/255, 9/137, 10/83
    # Размерность дерева по умолчанию
    LEVELS = 6
    CHILDREN = 1624

    # id корня дерева 6/1624
    ROOT_ID = -9206407546997070313

    id = models.BigIntegerField(primary_key=True)
    parent = models.ForeignKey('self')
    
    class Meta:
        abstract = True



import os
from setuptools import setup
import treensl

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-treensl',
    version=treensl.__version__,
    packages=['treensl', 'treensl.migrations'],
    include_package_data=True,
    license='BSD License',
    description='Django app to tree in DB.',
    long_description=README,
    url='https://github.com/EvgeniyBurdin/django_treensl',
    author='Evgeniy Burdin',
    author_email='e.s.burdin@mail.ru',
    keywords='django',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',

    ],
)